# -*- encoding: utf-8 -*-
import time

import contextlib2
import datetime
import dateutil
import io
import json
import mock
import pytest
import responses
import threading
import trafaret as t

from datarobot import BatchPredictionJob, Credential


@pytest.fixture
def batch_prediction_jobs_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "count": 4,
        "previous": None,
        "next": None,
        "data": [{
            "status": "INITIALIZING",
            "percentageCompleted": 0,
            "elapsedTimeSec": 7747,
            "links": {
                "self": (
                    "https://host_name.com/batchPredictions/"
                    "5ce1204b962d741661907ea0/",
                ),
                "csvUpload": (
                    "https://host_name.com/batchPredictions/"
                    "5ce1204b962d741661907ea0/csvUpload/"
                ),
            },
            "jobSpec": {
                "numConcurrent": 1,
                "thresholdHigh": None,
                "thresholdLow": None,
                "filename": "",
                "deploymentId": "5ce1138c962d7415e076d8c6",
                "passthroughColumns": [],
                "passthroughColumnsSet": None,
                "maxExplanations": None,
            },
            "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
        }, {
            "id": "5ce1204b962d741661907ea0",
            "status": "INITIALIZING",
            "percentageCompleted": 0,
            "elapsedTimeSec": 7220,
            "links": {
                "self": (
                    "https://host_name.com/batchPredictions/"
                    "5ce1225a962d741661907eb3/",
                )
            },
            "jobSpec": {
                "numConcurrent": 1,
                "thresholdHigh": None,
                "thresholdLow": None,
                "filename": "",
                "deploymentId": "5ce1138c962d7415e076d8c6",
                "passthroughColumns": [],
                "passthroughColumnsSet": None,
                "maxExplanations": None
            },
            "statusDetails": "Job submitted at 2019-05-19 09:31:06.724000"
        }]
    })


@pytest.fixture
def batch_prediction_job_initializing_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "INITIALIZING",
        "percentageCompleted": 0,
        "elapsedTimeSec": 7747,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
            "csvUpload": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/csvUpload/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_s3_initializing_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "INITIALIZING",
        "percentageCompleted": 0,
        "elapsedTimeSec": 7747,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {"type": "s3", "url": "s3://bucket/source_key"},
            "output_settings": {"type": "s3", "url": "s3://bucket/target_key"}
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_azure_initializing_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "INITIALIZING",
        "percentageCompleted": 0,
        "elapsedTimeSec": 7747,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {
                "type": "azure",
                "url": "https://storage_account.blob.endpoint/container/source_key"
            },
            "output_settings": {
                "type": "azure",
                "url": "https://storage_account.blob.endpoint/container/target_key"
            }
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_gcp_initializing_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "INITIALIZING",
        "percentageCompleted": 0,
        "elapsedTimeSec": 7747,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {
                "type": "gcp",
                "url": "https://storage.googleapis.com/bucket/source_key"
            },
            "output_settings": {
                "type": "gcp",
                "url": "https://storage.googleapis.com/bucket/target_key"
            }
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_running_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "RUNNING",
        "scored_rows": 400,
        "failed_rows": 0,
        "percentageCompleted": 40,
        "elapsedTimeSec": 7747,
        "links": {
            "download": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/download/"
            ),
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_completed_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 400,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 7747,
        "links": {
            "download": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/download/"
            ),
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_completed_passthrough_columns_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 400,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 7747,
        "links": {
            "download": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/download/"
            ),
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": ["a", "b"],
            "passthroughColumnsSet": None,
            "maxExplanations": None
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_completed_passthrough_columns_set_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 400,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 7747,
        "links": {
            "download": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/download/"
            ),
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": "all",
            "maxExplanations": None
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_s3_completed_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 400,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 7747,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {"type": "s3", "url": "s3://bucket/source_key"},
            "output_settings": {"type": "s3", "url": "s3://bucket/target_key"}
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_azure_completed_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 300,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 5561,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {
                "type": "azure",
                "url": "https://storage_account.blob.endpoint/container/source_key"
            },
            "output_settings": {
                "type": "azure",
                "url": "https://storage_account.blob.endpoint/container/target_key"
            }
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_gcp_completed_json():
    return json.dumps({
        "id": "5ce1204b962d741661907ea0",
        "status": "COMPLETED",
        "scored_rows": 500,
        "failed_rows": 0,
        "percentageCompleted": 100,
        "elapsedTimeSec": 1451,
        "links": {
            "self": (
                "https://host_name.com/batchPredictions/"
                "5ce1204b962d741661907ea0/"
            ),
        },
        "jobSpec": {
            "numConcurrent": 1,
            "thresholdHigh": None,
            "thresholdLow": None,
            "filename": "",
            "deploymentId": "5ce1138c962d7415e076d8c6",
            "passthroughColumns": [],
            "passthroughColumnsSet": None,
            "maxExplanations": None,
            "intake_settings": {
                "type": "gcp",
                "url": "https://storage.googleapis.com/bucket/source_key"
            },
            "output_settings": {
                "type": "gcp",
                "url": "https://storage.googleapis.com/bucket/target_key"
            }
        },
        "statusDetails": "Job submitted at 2019-05-19 09:22:19.779000"
    })


@pytest.fixture
def batch_prediction_job_data_csv():
    return b"""readmitted_1.0_PREDICTION,readmitted_0.0_PREDICTION,readmitted_PREDICTION,THRESHOLD,POSITIVE_CLASS,prediction_status
0.219181314111,0.780818685889,0.0,0.5,1.0,OK
0.341459780931,0.658540219069,0.0,0.5,1.0,OK
0.420107662678,0.579892337322,0.0,0.5,1.0,OK"""


@responses.activate
@pytest.mark.usefixtures('client')
def test_list_batch_prediction_jobs_by_status(batch_prediction_jobs_json):
    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/',
        body=batch_prediction_jobs_json)

    job_statuses = BatchPredictionJob.list_by_status()

    assert 2 == len(job_statuses)


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize(
    ['job_fixture', 'expected_status', 'expected_percentage_completed'],
    [
        pytest.param(
            'batch_prediction_job_initializing_json',
            'INITIALIZING',
            0,
        ),
        pytest.param(
            'batch_prediction_job_completed_passthrough_columns_json',
            'COMPLETED',
            100,
        ),
        pytest.param(
            'batch_prediction_job_completed_passthrough_columns_set_json',
            'COMPLETED',
            100,
        ),
        pytest.param(
            'batch_prediction_job_s3_completed_json',
            'COMPLETED',
            100,
        ),
    ]
)
def test_get_batch_prediction_job_status(
        request, job_fixture, expected_status, expected_percentage_completed):

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/',
        body=request.getfixturevalue(job_fixture)
    )

    job = BatchPredictionJob.get('5ce1204b962d741661907ea0')
    job_status = job.get_status()

    assert job_status['status'] == expected_status
    assert job_status['percentage_completed'] == expected_percentage_completed


@responses.activate
@pytest.mark.usefixtures('client')
def test_get_result_when_done(
        batch_prediction_job_initializing_json,
        batch_prediction_job_running_json,
        batch_prediction_job_completed_json,
        batch_prediction_job_data_csv):

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/',
        body=batch_prediction_job_initializing_json)

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/',
        body=batch_prediction_job_running_json)

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/',
        body=batch_prediction_job_completed_json)

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/',
        body=batch_prediction_job_data_csv)

    job = BatchPredictionJob.get('5ce1204b962d741661907ea0')

    assert job.get_result_when_complete() == batch_prediction_job_data_csv


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize(['kwargs', 'expected_read_timeout'], [
    pytest.param({}, 660, id='default-timeout'),
    pytest.param({'download_read_timeout': 200}, 200, id='override-timeout'),
])
def test_score_to_file(
    tmpdir,
    kwargs,
    expected_read_timeout,
    batch_prediction_job_initializing_json,
    batch_prediction_job_completed_json,
    batch_prediction_job_data_csv
):

    responses.add(
        responses.POST,
        "https://host_name.com/batchPredictions/",
        body=batch_prediction_job_initializing_json,
        headers={
            "Location": "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        },
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/",
        body=batch_prediction_job_data_csv,
    )

    responses.add(
        responses.PUT,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/csvUpload/",
        body="",
        status=202,
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body=batch_prediction_job_completed_json,
    )

    output_file = str(tmpdir.mkdir("sub").join("scored.csv"))
    thread_count_before = threading.activeCount()

    with mock.patch.object(BatchPredictionJob._client, 'get',
                           wraps=BatchPredictionJob._client.get) as download_spy:

        BatchPredictionJob.score_to_file(
            "5ce1138c962d7415e076d8c6",
            io.BytesIO(b"foo\nbar"),
            output_file,
            **kwargs
        )

        download_spy.assert_any_call(
            'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/',
            stream=True,
            timeout=expected_read_timeout,
        )

    assert open(output_file, "rb").read() == batch_prediction_job_data_csv
    assert thread_count_before == threading.activeCount(), 'Thread leak'


@responses.activate
@pytest.mark.usefixtures('client')
def test_score_to_file_timeout(
    tmpdir,
    batch_prediction_job_initializing_json,
    batch_prediction_job_completed_json,
    batch_prediction_job_data_csv
):

    responses.add(
        responses.POST,
        "https://host_name.com/batchPredictions/",
        body=batch_prediction_job_initializing_json,
        headers={
            "Location": "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        },
    )

    responses.add(
        responses.PUT,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/csvUpload/",
        body="",
        status=202,
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body=batch_prediction_job_initializing_json,
    )

    responses.add(
        responses.DELETE,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body="",
        status=202,
    )

    output_file = str(tmpdir.mkdir("sub").join("scored.csv"))

    thread_count_before = threading.activeCount()

    with pytest.raises(RuntimeError):
        BatchPredictionJob.score_to_file(
            "5ce1138c962d7415e076d8c6",
            io.BytesIO(b"foo\nbar"),
            output_file,
            download_timeout=1,
        )

    assert open(output_file, 'r').read() == ''
    assert thread_count_before == threading.activeCount(), 'Thread leak'

    # Job should abort after timeout
    last_request = responses.calls[-1].request
    assert last_request.method == 'DELETE'
    assert last_request.url == 'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/'


@responses.activate
@pytest.mark.usefixtures('client')
@mock.patch('requests.sessions.Session.put')
def test_score_to_file_race_condition(
    put_mock,
    tmpdir,
    batch_prediction_job_initializing_json,
    batch_prediction_job_completed_json,
    batch_prediction_job_data_csv
):
    responses.add(
        responses.POST,
        "https://host_name.com/batchPredictions/",
        body=batch_prediction_job_initializing_json,
        headers={
            "Location": "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        },
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/",
        body=batch_prediction_job_data_csv,
    )

    responses.add(
        responses.PUT,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/csvUpload/",
        body="",
        status=202,
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body=batch_prediction_job_completed_json,
    )

    output_file = str(tmpdir.mkdir("sub").join("scored.csv"))

    thread_count_before = threading.activeCount()

    # It is not possible to slow down the PUT request itself, as the bug this test fixes
    # was related to a passed argument, which is already sent on the stack at this point.
    # Therefore, we slow down the seek(0) operation happening immediately before, triggering
    # the race condition.
    def slow_seek(*args):
        time.sleep(1)
        return

    mock_file_object = mock.MagicMock()
    mock_file_object.seek.side_effect = slow_seek

    BatchPredictionJob.score_to_file(
        "5ce1138c962d7415e076d8c6",
        mock_file_object,
        output_file,
    )

    put_mock.assert_called_once()
    assert open(output_file, "rb").read() == batch_prediction_job_data_csv
    assert thread_count_before == threading.activeCount(), 'Thread leak'


@responses.activate
@pytest.mark.usefixtures('client')
@pytest.mark.parametrize([
    'name',
    'scoring_function',
    'source_url',
    'destination_url'
], [
    pytest.param(
        's3',
        BatchPredictionJob.score_s3,
        's3://bucket/source_key',
        's3://bucket/target_key',
        id='s3',
    ),
    pytest.param(
        'azure',
        BatchPredictionJob.score_azure,
        'https://storage_account.blob.endpoint/container/source_key',
        'https://storage_account.blob.endpoint/container/target_key',
        id='azure',
    ),
    pytest.param(
        'gcp',
        BatchPredictionJob.score_gcp,
        'https://storage.googleapis.com/bucket/source_key',
        'https://storage.googleapis.com/bucket/target_key',
        id='gcp',
    ),
])
def test_score_cloud(
    request,
    name,
    scoring_function,
    source_url,
    destination_url
):
    responses.add(
        responses.POST,
        "https://host_name.com/batchPredictions/",
        body=request.getfixturevalue(
            'batch_prediction_job_{name}_initializing_json'.format(name=name)
        ),
        headers={
            "Location": "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        },
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body=request.getfixturevalue(
            'batch_prediction_job_{name}_completed_json'.format(name=name)
        ),
    )

    job = scoring_function(
        deployment="5ce1138c962d7415e076d8c6",
        source_url=source_url,
        destination_url=destination_url,
        credential=Credential('key_id'),
    )

    job.wait_for_completion()
    job_status = job.get_status()

    assert job_status['job_spec']['intake_settings']['type'] == name
    assert job_status['job_spec']['output_settings']['type'] == name
    assert job_status['job_spec']['intake_settings']['url'] == source_url
    assert job_status['job_spec']['output_settings']['url'] == destination_url


@responses.activate
@pytest.mark.usefixtures('client')
def test_score_from_existing(batch_prediction_job_s3_completed_json):

    responses.add(
        responses.POST,
        "https://host_name.com/batchPredictions/fromExisting/",
        headers={
            "Location": "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        },
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/",
        body=batch_prediction_job_s3_completed_json,
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea1/",
        body=batch_prediction_job_s3_completed_json,
    )

    job = BatchPredictionJob.score_from_existing(
        batch_prediction_job_id="5ce1204b962d741661907ea1",
    )

    job.wait_for_completion()


@pytest.mark.usefixtures('client')
@pytest.mark.parametrize([
    'score_args',
    'expected_exception',
    'expected_message'
], [
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 'unknown'}
        },
        ValueError,
        'Unsupported type parameter for intake_settings',
        id='unsupported-intake-option',
    ),
    pytest.param(
        {
            'deployment': 'foo',
        },
        ValueError,
        'Missing source data',
        id='missing-source-data',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 's3'}
        },
        t.DataError,
        None,
        id='missing-s3-intake-configuration',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 's3', 'url': 's3://bucket/source_key'},
            'output_settings': {'type': 's3'}
        },
        t.DataError,
        None,
        id='missing-s3-output-configuration',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 'local_file'},
            'output_settings': {'type': 'local_file'},
            'timeseries_settings': {'type': 'unknown'},
        },
        t.DataError,
        None,
        id='unknown-ts-prediction-type',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 'local_file'},
            'output_settings': {'type': 'local_file'},
            'timeseries_settings': {
                'type': 'historical',
                'forecast_point': '2020-05-16T17:42:12+00:00',
            },
        },
        t.DataError,
        None,
        id='forecast-point-for-ts-historical-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 'local_file'},
            'output_settings': {'type': 'local_file'},
            'timeseries_settings': {
                'type': 'forecast',
                'predictions_start_date': '2020-05-16T17:42:12+00:00',
            },
        },
        t.DataError,
        None,
        id='predictions-start-date-for-ts-forecast-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': '2020-05-16T17:42:13+00:00',
            },
        },
        ValueError,
        None,
        id='predictions-start-date-without-end-date-for-ts-historical-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': '2020-05-17T17:42:13+00:00',
                'predictions_end_date': '2020-05-16T17:42:13+00:00',
            },
        },
        ValueError,
        None,
        id='predictions-start-date-after-end-date-for-ts-historical-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'forecast',
                'forecast_point': 'foo',
            },
        },
        ValueError,
        None,
        id='forecast-point-with-wrong-format-for-ts-forecast-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': 'foo',
                'predictions_end_date': '2020-05-17T17:42:13+00:00',
            },
        },
        ValueError,
        None,
        id='predictions-start-date-with-wrong-format-for-ts-historical-predictions',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': '2020-05-16T17:42:13+00:00',
                'predictions_end_date': 'foo',
            },
        },
        ValueError,
        None,
        id='predictions-end-date-with-wrong-format-for-ts-historical-predictions',
    ),
])
def test_score_errors(score_args, expected_exception, expected_message):
    with contextlib2.ExitStack() as stack:

        if 'timeseries_settings' in score_args:
            stack.enter_context(pytest.warns(UserWarning))

        stack.enter_context(pytest.raises(expected_exception, match=expected_message))

        BatchPredictionJob.score(**score_args)


@pytest.mark.parametrize([
    'score_args',
    'expected_job_data',
], [
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
        },
        id='deployment-id',
    ),
    pytest.param(
        {
            'deployment': mock.MagicMock(id='bar'),
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
        },
        {
            'deploymentId': 'bar',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
        },
        id='deployment',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 's3', 'url': 's3://bucket/source_key'},
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 's3', 'url': 's3://bucket/source_key'},
            'outputSettings': {'type': 'localFile'},
        },
        id='s3-intake-settings',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {'type': 's3', 'url': 's3://bucket/source_key'},
            'output_settings': {'type': 's3', 'url': 's3://bucket/target_key'},
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 's3', 'url': 's3://bucket/source_key'},
            'outputSettings': {'type': 's3', 'url': 's3://bucket/target_key'},
        },
        id='s3-intake-output-settings',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 's3',
                'url': 's3://bucket/source_key',
                'credential_id': 'key_id',
            },
            'output_settings': {
                'type': 's3',
                'url': 's3://bucket/target_key',
                'credential_id': 'key_id',
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {
                'type': 's3',
                'url': 's3://bucket/source_key',
                'credentialId': 'key_id',
            },
            'outputSettings': {
                'type': 's3',
                'url': 's3://bucket/target_key',
                'credentialId': 'key_id',
            },
        },
        id='full-s3-intake-output-settings',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'threshold_high': 0.95,
            'threshold_low': 0.05,
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'thresholdHigh': 0.95,
            'thresholdLow': 0.05,
        },
        id='thresholds',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'passthrough_columns': ['a', 'b', 'c'],
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'passthroughColumns': ['a', 'b', 'c'],
        },
        id='passthrough-columns',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'passthrough_columns_set': 'all',
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'passthroughColumnsSet': 'all',
        },
        id='passthrough-columns-set-all',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'passthrough_columns': ['a', 'b', 'c'],
            'passthrough_columns_set': 'all',
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'passthroughColumnsSet': 'all',
        },
        id='passthrough-columns-set-override',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'num_concurrent': 10
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'numConcurrent': 10
        },
        id='num-concurrent',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'max_explanations': 10
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'maxExplanations': 10
        },
        id='max-explanations',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'prediction_warning_enabled': True
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'predictionWarningEnabled': True,
        },
        id='prediction-warning-enabled',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'jdbc',
                'table': 'test',
                'schema': 'public',
                'data_store_id': 'abcd1234',
                'credential_id': 'key_id',
            },
            'output_settings': {
                'type': 'jdbc',
                'table': 'test2',
                'schema': 'public',
                'statement_type': 'insert',
                'data_store_id': 'abcd1234',
                'credential_id': 'key_id',
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {
                'type': 'jdbc',
                'table': 'test',
                'schema': 'public',
                'dataStoreId': 'abcd1234',
                'credentialId': 'key_id',
            },
            'outputSettings': {
                'type': 'jdbc',
                'table': 'test2',
                'schema': 'public',
                'statementType': 'insert',
                'dataStoreId': 'abcd1234',
                'credentialId': 'key_id',
            },
        },
        id='full-jdbc-intake-output-settings',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {'type': 'forecast'}
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {'type': 'forecast'},
        },
        id='ts-forecast-default-forecast-point',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'forecast',
                'forecast_point': '2020-05-16T17:42:13+00:00',
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {
                'type': 'forecast',
                'forecastPoint': '2020-05-16T17:42:13+00:00',
            },
        },
        id='ts-forecast-string-forecast-point',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'forecast',
                'forecast_point': datetime.datetime(
                    2020, 5, 16, 17, 42, 13, tzinfo=dateutil.tz.tzutc()
                ),
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {
                'type': 'forecast',
                'forecastPoint': '2020-05-16T17:42:13+00:00',
            },
        },
        id='ts-forecast-datetime-forecast-point',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {'type': 'historical'},
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {'type': 'historical'},
        },
        id='ts-historical-default',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': '2020-05-16T17:42:13+00:00',
                'predictions_end_date': '2020-05-17T17:42:13+00:00',
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {
                'type': 'historical',
                'predictionsStartDate': '2020-05-16T17:42:13+00:00',
                'predictionsEndDate':  '2020-05-17T17:42:13+00:00',
            },
        },
        id='ts-historical-string-prediction-interval',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'predictions_start_date': datetime.datetime(
                    2020, 5, 16, 17, 42, 13, tzinfo=dateutil.tz.tzutc()
                ),
                'predictions_end_date': datetime.datetime(
                    2020, 5, 17, 17, 42, 13, tzinfo=dateutil.tz.tzutc()
                ),
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {
                'type': 'historical',
                'predictionsStartDate': '2020-05-16T17:42:13+00:00',
                'predictionsEndDate': '2020-05-17T17:42:13+00:00',
            },
        },
        id='ts-historical-datetime-prediction-interval',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'timeseries_settings': {
                'type': 'historical',
                'relax_known_in_advance_features_check': True,
            },
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'timeseriesSettings': {
                'type': 'historical',
                'relaxKnownInAdvanceFeaturesCheck': True,
            },
        },
        id='ts-historical-default-relax-kia',
    ),
    pytest.param(
        {
            'deployment': 'foo',
            'intake_settings': {
                'type': 'localFile',
                'file': io.BytesIO(b'foo\nbar'),
            },
            'prediction_instance': {
                'hostName': '192.0.2.4',
                'sslEnabled': False,
                'apiKey': 'NWUQ9w21UhGgerBtOC4ahN0aqjbjZ0NMhL1e5cSt4ZHIBn2w',
                'datarobotKey': '154a8abb-cbde-4e73-ab3b-a46c389c337b'
            }
        },
        {
            'deploymentId': 'foo',
            'intakeSettings': {'type': 'localFile'},
            'outputSettings': {'type': 'localFile'},
            'predictionInstance': {
                'hostName': '192.0.2.4',
                'sslEnabled': False,
                'apiKey': 'NWUQ9w21UhGgerBtOC4ahN0aqjbjZ0NMhL1e5cSt4ZHIBn2w',
                'datarobotKey': '154a8abb-cbde-4e73-ab3b-a46c389c337b'
            }
        },
        id='prediction-instance',
    ),
])
@mock.patch('datarobot.BatchPredictionJob._client')
def test_score_job_data(mock_client, score_args, expected_job_data):

    # Using an exception to short-circuit the score function and test
    # the contents of the job data without proceeding with the rest
    # of the function

    mock_client.post.side_effect = RuntimeError('short-circuit')

    with contextlib2.ExitStack() as stack:

        if 'timeseries_settings' in score_args:
            stack.enter_context(pytest.warns(UserWarning))

        stack.enter_context(pytest.raises(RuntimeError, match='short-circuit'))

        BatchPredictionJob.score(**score_args)

    mock_client.post.assert_called_once_with(
        url=BatchPredictionJob._jobs_path(),
        json=expected_job_data,
    )


@pytest.mark.parametrize(['download_kwargs', 'expected_read_timeout'], [
    pytest.param({}, 660, id='default-timeout'),
    pytest.param({'read_timeout': 200}, 200, id='override-timeout'),
])
@responses.activate
def test_download_read_timeout(
    batch_prediction_job_completed_json,
    batch_prediction_job_data_csv,
    download_kwargs,
    expected_read_timeout
):

    job_id = '5ce1204b962d741661907ea0'

    responses.add(
        responses.GET,
        'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/',
        body=batch_prediction_job_completed_json
    )

    responses.add(
        responses.GET,
        "https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/",
        body=batch_prediction_job_data_csv,
    )

    job = BatchPredictionJob.get(job_id)
    buf = io.BytesIO()

    with mock.patch.object(BatchPredictionJob._client, 'get',
                           wraps=BatchPredictionJob._client.get) as download_spy:

        job.download(buf, **download_kwargs)

        download_spy.assert_any_call(
            'https://host_name.com/batchPredictions/5ce1204b962d741661907ea0/download/',
            stream=True,
            timeout=expected_read_timeout,
        )



# coding: utf-8
import json
import pytest
import responses

from trafaret import DataError
from datarobot.errors import ClientError, IllegalFileName

from datarobot import CalendarFile, Project, SharingAccess, enums
from tests.test_helpers import fixture_file_path

_base_url = 'https://host_name.com/calendars/'
_valid_calendar_id = 'calId'
_invalid_calendar_id = 'badCalId'
_valid_calendar_url = '{0}{1}/'.format(_base_url, _valid_calendar_id)
_invalid_calendar_url = '{0}{1}/'.format(_base_url, _invalid_calendar_id)

calendar_dict = {
    'numEvents': 5,
    'name': 'calendar.csv',
    'calendarStartDate': '2018-01-01',
    'created': '2018-07-16T17:28:13.929687Z',
    'numEventTypes': 5,
    'source': 'calendar.csv',
    'calendarEndDate': '2018-07-04',
    'projectIds': [],
    'Id': 'calId',
    'role': 'OWNER',
    'multiseriesIdColumns': None,
}

calendar_dict_with_name = {
    'numEvents': 5,
    'name': 'My Calendar',
    'calendarStartDate': '2018-01-01',
    'created': '2018-07-16T17:28:13.929687Z',
    'numEventTypes': 5,
    'source': 'calendar.csv',
    'calendarEndDate': '2018-07-04',
    'projectIds': [],
    'Id': 'calId',
    'role': 'OWNER',
    'multiseriesIdColumns': None,
}


multiseries_calendar_dict = {
    'numEvents': 32,
    'name': 'multiseries_calendar.csv',
    'calendarStartDate': '2019-01-01',
    'created': '2019-10-17T17:28:13.929687Z',
    'numEventTypes': 5,
    'source': 'multiseries_calendar.csv',
    'calendarEndDate': '2019-12-26',
    'projectIds': [],
    'Id': 'calId',
    'role': 'OWNER',
    'multiseriesIdColumns': ['series'],
}


multiseries_calendar_dict_with_name = {
    'numEvents': 32,
    'name': 'My Multiseries Calendar',
    'calendarStartDate': '2019-01-01',
    'created': '2019-10-17T17:28:13.929687Z',
    'numEventTypes': 5,
    'source': 'multiseries_calendar.csv',
    'calendarEndDate': '2019-12-26',
    'projectIds': [],
    'Id': 'calId',
    'role': 'OWNER',
    'multiseriesIdColumns': ['series'],
}


@pytest.fixture
def calendar_json():
    return json.dumps(calendar_dict)


@pytest.fixture
def calendar_with_name_json():
    return json.dumps(calendar_dict_with_name)


@pytest.fixture
def multiseries_calendar_json():
    return json.dumps(multiseries_calendar_dict)


@pytest.fixture
def multiseries_calendar_with_name_json():
    return json.dumps(multiseries_calendar_dict_with_name)


@pytest.fixture
def calendar_list_json():
    cal_list = []
    for i in range(1, 11):
        cal_copy = calendar_dict.copy()
        cal_copy['name'] = 'cal_name{0}'.format(i)
        cal_list.append(cal_copy)
    return json.dumps({
        'count': len(cal_list),
        'data': cal_list,
        'next': None,
        'previous': None
    })


@pytest.fixture
def multiseries_calendar_list_json():
    cal_list = []
    for i in range(1, 11):
        cal_copy = multiseries_calendar_dict.copy()
        cal_copy['name'] = 'cal_name{0}'.format(i)
        cal_list.append(cal_copy)
    return json.dumps({
        'count': len(cal_list),
        'data': cal_list,
        'next': None,
        'previous': None
    })


@pytest.fixture
def calendar_access_control_list_json():
    ac_list = [{'userId': 'ac{0}'.format(i), 'username': 'user{0}'.format(i),
                'canShare': True, 'role': 'OWNER'} for i in range(1, 11)]
    return json.dumps({
        'count': len(ac_list),
        'data': ac_list,
        'next': None,
        'previous': None
    })


@pytest.fixture
def calendar_invalid_json():
    return json.dumps({"message": "Not Found"})


@responses.activate
def test_calendar_create(calendar_json):
    responses.add(
        responses.POST,
        _base_url + 'fileUpload/',
        body='',
        status=202,
        adding_headers={'Location': 'https://host_name.com/status/some-status/'},
        content_type='application_json'
    )
    responses.add(
        responses.GET,
        'https://host_name.com/status/some-status/',
        body='',
        status=303,
        adding_headers={'Location': _valid_calendar_url},
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        _valid_calendar_url,
        body=calendar_json,
        status=200, content_type='application/json'
    )
    calendar = CalendarFile.create(fixture_file_path('calendar.csv'))
    assert isinstance(calendar, CalendarFile)
    assert calendar.id == 'calId'
    assert calendar.num_events == 5
    assert calendar.name == 'calendar.csv'
    assert calendar.calendar_start_date == '2018-01-01'
    assert calendar.calendar_end_date == '2018-07-04'
    assert calendar.created == '2018-07-16T17:28:13.929687Z'
    assert calendar.num_event_types == 5
    assert calendar.project_ids == []
    assert calendar.role == 'OWNER'
    assert not calendar.multiseries_id_columns


def _test_multiseries_calendar_create(multiseries_calendar_json, name=None):
    responses.add(
        responses.POST,
        _base_url + 'fileUpload/',
        body='',
        status=202,
        adding_headers={'Location': 'https://host_name.com/status/some-status/'},
        content_type='application_json'
    )
    responses.add(
        responses.GET,
        'https://host_name.com/status/some-status/',
        body='',
        status=303,
        adding_headers={'Location': _valid_calendar_url},
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        _valid_calendar_url,
        body=multiseries_calendar_json,
        status=200, content_type='application/json'
    )
    calendar = CalendarFile.create(
        fixture_file_path('multiseries_calendar.csv'), calendar_name=name
    )
    assert isinstance(calendar, CalendarFile)
    assert calendar.id == 'calId'
    assert calendar.num_events == 32
    assert calendar.name == name if name else 'multiseries_calendar.csv'
    assert calendar.calendar_start_date == '2019-01-01'
    assert calendar.calendar_end_date == '2019-12-26'
    assert calendar.created == '2019-10-17T17:28:13.929687Z'
    assert calendar.num_event_types == 5
    assert calendar.project_ids == []
    assert calendar.role == 'OWNER'
    assert calendar.multiseries_id_columns == ['series']


@responses.activate
def test_multiseries_calendar_create(multiseries_calendar_json):
    _test_multiseries_calendar_create(multiseries_calendar_json)


@responses.activate
def test_multiseries_calendar_create_with_name(multiseries_calendar_with_name_json):
    _test_multiseries_calendar_create(
        multiseries_calendar_with_name_json, name='My Multiseries Calendar'
    )


def test_calendar_create_non_existent_file():
    with pytest.raises(ValueError):
        CalendarFile.create('bad_file_path')


def test_calendar_create_not_a_string():
    with pytest.raises(ValueError):
        CalendarFile.create({'this definitely': 'isn"t a string'})


def test_calendar_create_multiseries_id_not_a_list():
    with pytest.raises(ValueError) as excinfo:
        CalendarFile.create(
            fixture_file_path('multiseries_calendar.csv'),
            multiseries_id_columns='bad_id'
        )
    assert str(excinfo.value) == 'Expected list of str for multiseries_id_columns, got: bad_id'


@pytest.fixture
def project_without_partition(project_without_target_data):
    project_without_target_data['partition'] = None
    return Project(**project_without_target_data)


@responses.activate
def test_calendar_create_with_name(calendar_with_name_json):
    responses.add(
        responses.POST,
        _base_url + 'fileUpload/',
        body='',
        status=202,
        adding_headers={'Location': 'https://host_name.com/status/some-status/'},
        content_type='application_json'
    )
    responses.add(
        responses.GET,
        'https://host_name.com/status/some-status/',
        body='',
        status=303,
        adding_headers={'Location': _valid_calendar_url},
        content_type='application/json'
    )
    responses.add(
        responses.GET,
        _valid_calendar_url,
        body=calendar_with_name_json,
        status=200, content_type='application/json'
    )
    calendar = CalendarFile.create(fixture_file_path('calendar.csv'),
                                   calendar_name='My Calendar')
    assert isinstance(calendar, CalendarFile)
    assert calendar.id == 'calId'
    assert calendar.num_events == 5
    assert calendar.name == 'My Calendar'
    assert calendar.calendar_start_date == '2018-01-01'
    assert calendar.calendar_end_date == '2018-07-04'
    assert calendar.created == '2018-07-16T17:28:13.929687Z'
    assert calendar.num_event_types == 5
    assert calendar.project_ids == []
    assert calendar.role == 'OWNER'


@responses.activate
def test_calendar_get(calendar_json):
    responses.add(
        responses.GET,
        _valid_calendar_url,
        body=calendar_json,
        status=200,
        content_type='application/json'
    )
    calendar = CalendarFile.get(_valid_calendar_id)
    assert isinstance(calendar, CalendarFile)
    assert calendar.id == 'calId'
    assert calendar.num_events == 5
    assert calendar.name == 'calendar.csv'
    assert calendar.calendar_start_date == '2018-01-01'
    assert calendar.calendar_end_date == '2018-07-04'
    assert calendar.created == '2018-07-16T17:28:13.929687Z'
    assert calendar.num_event_types == 5
    assert calendar.project_ids == []
    assert calendar.role == 'OWNER'


@responses.activate
def test_calendar_invalid_get(calendar_invalid_json):
    responses.add(responses.GET,
                  _invalid_calendar_url,
                  body=calendar_invalid_json)
    with pytest.raises(DataError):
        CalendarFile.get(_invalid_calendar_id)


@responses.activate
def test_calendar_list(calendar_list_json):
    responses.add(responses.GET,
                  _base_url,
                  body=calendar_list_json)
    cal_list = CalendarFile.list()
    assert len(cal_list) == 10
    for item in cal_list:
        assert isinstance(item, CalendarFile)


@responses.activate
def test_multiseries_calendar_list(multiseries_calendar_list_json):
    responses.add(responses.GET,
                  _base_url,
                  body=multiseries_calendar_list_json)
    cal_list = CalendarFile.list()
    assert len(cal_list) == 10
    for item in cal_list:
        assert isinstance(item, CalendarFile)


@responses.activate
def test_calendar_delete_valid():
    responses.add(responses.DELETE,
                  _valid_calendar_url,
                  status=204)
    assert CalendarFile.delete('calId') is None


@responses.activate
def test_calendar_delete_invalid():
    responses.add(responses.DELETE,
                  _invalid_calendar_url,
                  status=404)
    with pytest.raises(ClientError):
        CalendarFile.delete(_invalid_calendar_id)


@responses.activate
def test_calendar_update_name_valid():
    responses.add(responses.PATCH,
                  _valid_calendar_url,
                  status=200)
    assert CalendarFile.update_name(_valid_calendar_id, 'new name') == 200


@responses.activate
def test_calendar_update_name_invalid_calendar():
    responses.add(responses.PATCH,
                  _invalid_calendar_url,
                  status=404)
    with pytest.raises(ClientError):
        CalendarFile.update_name(_invalid_calendar_id, 'new name')


@responses.activate
def test_calendar_update_name_invalid_name():
    with pytest.raises(IllegalFileName):
        CalendarFile.update_name(_valid_calendar_id, '£ ¥\n§ ¾')


@responses.activate
def test_calendar_sharing_valid():
    responses.add(responses.PATCH,
                  _valid_calendar_url + 'accessControl/',
                  status=200)
    payload = [SharingAccess('valid_username@datarobot.com', enums.SHARING_ROLE.READ_WRITE)]
    assert CalendarFile.share(_valid_calendar_id, payload) == 200


@responses.activate
def test_calendar_sharing_invalid_username():
    responses.add(responses.PATCH,
                  _invalid_calendar_url + 'accessControl/',
                  status=422)
    payload = [SharingAccess('invalid_valid_username@datarobot.com', enums.SHARING_ROLE.READ_WRITE)]
    with pytest.raises(ClientError):
        CalendarFile.share(_invalid_calendar_id, payload)


def test_calendar_sharing_invalid_access_list():
    # invalid access_list
    payload = {"look at me, I'm a": 'dict'}
    with pytest.raises(AssertionError):
        CalendarFile.share(_valid_calendar_id, payload)
    # invalid items in access_list
    payload = [i for i in range(10)]
    with pytest.raises(AssertionError):
        CalendarFile.share(_valid_calendar_id, payload)


@responses.activate
def test_calendar_sharing_access_list(calendar_access_control_list_json):
    responses.add(
        responses.GET,
        _valid_calendar_url + 'accessControl/',
        body=calendar_access_control_list_json,
    )
    access_control_list = CalendarFile.get_access_list(_valid_calendar_id)
    assert len(access_control_list) == 10
    for item in access_control_list:
        assert isinstance(item, SharingAccess)


@responses.activate
def test_calendar_sharing_access_list_user_no_access():
    responses.add(
        responses.GET,
        _valid_calendar_url + 'accessControl/',
        status=404
    )
    with pytest.raises(ClientError):
        CalendarFile.get_access_list(_valid_calendar_id)


def test_retrieve_calendar_id_project_pre_aim(project_without_partition):
    assert project_without_partition.calendar_id is None



import responses

from tests.utils import request_body_to_json
from datarobot.errors import InvalidUsageError


@responses.activate
def test_set_target(datetime_partition_spec, holdout_start_date,
                    project_with_target_json, project_without_target,
                    unittest_endpoint, project_id):
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in datetime_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': datetime_partition_spec.datetime_partition_column,
        'useTimeSeries': datetime_partition_spec.use_time_series,
        'defaultToKnownInAdvance': datetime_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': datetime_partition_spec.default_to_do_not_derive,
        'autopilotDataSelectionMethod': datetime_partition_spec.autopilot_data_selection_method,
        'validationDuration': datetime_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': datetime_partition_spec.holdout_duration,
        'gapDuration': datetime_partition_spec.gap_duration,
        'numberOfBacktests': datetime_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
        'modelSplits': 2
    }
    project_without_target.set_target('target', partitioning_method=datetime_partition_spec)
    assert responses.calls[0].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[0].request)

    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_time_series_set_target(time_series_partition_spec, holdout_start_date,
                                project_with_target_json, project_without_target,
                                unittest_endpoint, project_id):
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)

    time_series_partition_spec.calendar_id = project_id

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': time_series_partition_spec.default_to_do_not_derive,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
        'calendarId': time_series_partition_spec.calendar_id
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[0].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[0].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_irregular_time_series_set_target(irregular_time_series_partition_spec,
                                          holdout_start_date,
                                          project_without_target,
                                          project_with_target_json,
                                          unittest_endpoint,
                                          project_id):
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    spec = irregular_time_series_partition_spec

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')
    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in irregular_time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': spec.datetime_partition_column,
        'useTimeSeries': spec.use_time_series,
        'defaultToKnownInAdvance': spec.default_to_known_in_advance,
        'defaultToDoNotDerive': spec.default_to_do_not_derive,
        'featureDerivationWindowStart': spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': spec.feature_derivation_window_end,
        'forecastWindowStart': spec.forecast_window_start,
        'forecastWindowEnd': spec.forecast_window_end,
        'autopilotDataSelectionMethod': spec.autopilot_data_selection_method,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': spec.holdout_duration,
        'gapDuration': spec.gap_duration,
        'numberOfBacktests': spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
    }
    project_without_target.set_target('target', partitioning_method=spec)

    assert responses.calls[0].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[0].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_multiseries_set_target(time_series_partition_spec, holdout_start_date,
                                project_with_target_json, project_without_target,
                                unittest_endpoint, project_id):
    time_series_partition_spec.multiseries_id_columns = ['series_id']
    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)
    date_feat_json = {
        'projectId': project_id, 'name': time_series_partition_spec.datetime_partition_column,
        'id': 2, 'timeSeriesEligible': False, 'featureType': 'date',
        'timeSeriesEligibilityReason': 'notUnique', 'lowInformation': False, 'uniqueCount': 500
    }
    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json=date_feat_json)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
        } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'multiseriesIdColumns': time_series_partition_spec.multiseries_id_columns,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': time_series_partition_spec.default_to_do_not_derive,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[2].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[2].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_use_cross_series_features(time_series_partition_spec, holdout_start_date,
                                   project_with_target_json, project_without_target,
                                   unittest_endpoint, project_id):
    time_series_partition_spec.multiseries_id_columns = ['series_id']
    time_series_partition_spec.use_cross_series_features = True
    time_series_partition_spec.aggregation_type = 'total'

    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)
    date_feat_json = {
        'projectId': project_id, 'name': time_series_partition_spec.datetime_partition_column,
        'id': 2, 'timeSeriesEligible': False, 'featureType': 'date',
        'timeSeriesEligibilityReason': 'notUnique', 'lowInformation': False, 'uniqueCount': 500
    }
    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json=date_feat_json)

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')
    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
    } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'multiseriesIdColumns': time_series_partition_spec.multiseries_id_columns,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': time_series_partition_spec.default_to_do_not_derive,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
        'useCrossSeriesFeatures': True,
        'aggregationType': 'total'
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[2].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[2].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_use_cross_series_group_by_with_columns_validated(
        time_series_partition_spec, holdout_start_date, project_with_target_json,
        project_without_target, unittest_endpoint, project_id):

    time_series_partition_spec.multiseries_id_columns = ['series_id']
    time_series_partition_spec.use_cross_series_features = True
    time_series_partition_spec.aggregation_type = 'total'
    time_series_partition_spec.cross_series_group_by_columns = ['correct_location']

    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    responses.add(responses.GET, multi_prop_url, json=multi_prop_json)

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.datetime_partition_column,
                      'id': 2,
                      'timeSeriesEligible': False,
                      'featureType': 'date',
                      'timeSeriesEligibilityReason': 'notUnique',
                      'lowInformation': False, 'uniqueCount': 500
                  })

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.multiseries_id_columns[0]),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.multiseries_id_columns[0],
                      'id': 3,
                      'timeSeriesEligible': True,
                      'featureType': 'category',
                      'timeSeriesEligibilityReason': 'suitable',
                      'lowInformation': False,
                      'uniqueCount': 100
                  })

    responses.add(responses.GET, ('{}multiseriesIds/{}/crossSeriesProperties/?'
                                  'crossSeriesGroupByColumns={}').format(
        project_url, time_series_partition_spec.multiseries_id_columns[0],
        time_series_partition_spec.cross_series_group_by_columns[0]),
                  json={
                      'multiseriesId': time_series_partition_spec.multiseries_id_columns[0],
                      'crossSeriesGroupByColumns': [
                          {
                              'name': time_series_partition_spec.cross_series_group_by_columns[0],
                              'eligibility': 'suitable',
                              'isEligible': True
                          }
                      ]
                  })

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
    } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'multiseriesIdColumns': time_series_partition_spec.multiseries_id_columns,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': time_series_partition_spec.default_to_do_not_derive,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
        'useCrossSeriesFeatures': True,
        'aggregationType': 'total',
        'crossSeriesGroupByColumns': time_series_partition_spec.cross_series_group_by_columns
    }
    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[4].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[4].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_use_cross_series_group_by_with_columns_not_validated(
        time_series_partition_spec, project_with_target_json, project_without_target,
        holdout_start_date, unittest_endpoint, project_id):
    time_series_partition_spec.multiseries_id_columns = ['series_id']
    time_series_partition_spec.use_cross_series_features = True
    time_series_partition_spec.aggregation_type = 'total'
    time_series_partition_spec.cross_series_group_by_columns = ['location']

    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    cross_series_url = '{}/projects/{}/crossSeriesProperties/'.format(unittest_endpoint, project_id)

    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }

    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.datetime_partition_column,
                      'id': 2,
                      'timeSeriesEligible': False,
                      'featureType': 'date',
                      'timeSeriesEligibilityReason': 'notUnique',
                      'lowInformation': False, 'uniqueCount': 500
                  })

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.multiseries_id_columns[0]),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.multiseries_id_columns[0],
                      'id': 3,
                      'timeSeriesEligible': True,
                      'featureType': 'category',
                      'timeSeriesEligibilityReason': 'suitable',
                      'lowInformation': False,
                      'uniqueCount': 100
                  })

    responses.add(responses.GET, ('{}multiseriesIds/{}/crossSeriesProperties/?'
                                  'crossSeriesGroupByColumns={}').format(
        project_url, time_series_partition_spec.multiseries_id_columns[0],
        time_series_partition_spec.cross_series_group_by_columns[0]),
                  json={
                      'multiseriesId': time_series_partition_spec.multiseries_id_columns[0],
                      'crossSeriesGroupByColumns': [
                          {
                              'name': time_series_partition_spec.cross_series_group_by_columns[0],
                              'eligibility': 'notAnalyzed',
                              'isEligible': False
                          }
                      ]
                  })

    responses.add(responses.POST,
                  cross_series_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')

    responses.add(responses.GET, ('{}multiseriesIds/{}/crossSeriesProperties/?'
                                  'crossSeriesGroupByColumns={}').format(
        project_url, time_series_partition_spec.multiseries_id_columns[0],
        time_series_partition_spec.cross_series_group_by_columns[0]),
                  json={
                      'multiseriesId': time_series_partition_spec.multiseries_id_columns[0],
                      'crossSeriesGroupByColumns': [
                          {
                              'name': time_series_partition_spec.cross_series_group_by_columns[0],
                              'eligibility': 'suitable',
                              'isEligible': True
                          }
                      ]
                  })

    expected_backtests = [{
        'index': bt.index, 'gapDuration': bt.gap_duration,
        'validationStartDate': bt.validation_start_date.isoformat(),
        'validationDuration': bt.validation_duration
    } for bt in time_series_partition_spec.backtests]
    expected_payload = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'multiseriesIdColumns': time_series_partition_spec.multiseries_id_columns,
        'useTimeSeries': time_series_partition_spec.use_time_series,
        'defaultToKnownInAdvance': time_series_partition_spec.default_to_known_in_advance,
        'defaultToDoNotDerive': time_series_partition_spec.default_to_do_not_derive,
        'featureDerivationWindowStart': time_series_partition_spec.feature_derivation_window_start,
        'featureDerivationWindowEnd': time_series_partition_spec.feature_derivation_window_end,
        'forecastWindowStart': time_series_partition_spec.forecast_window_start,
        'forecastWindowEnd': time_series_partition_spec.forecast_window_end,
        'autopilotDataSelectionMethod': time_series_partition_spec.autopilot_data_selection_method,
        'validationDuration': time_series_partition_spec.validation_duration,
        'holdoutStartDate': holdout_start_date.isoformat(),
        'holdoutDuration': time_series_partition_spec.holdout_duration,
        'gapDuration': time_series_partition_spec.gap_duration,
        'numberOfBacktests': time_series_partition_spec.number_of_backtests,
        'backtests': expected_backtests,
        'cvMethod': 'datetime',
        'target': 'target',
        'mode': 'auto',
        'quickrun': False,
        'useCrossSeriesFeatures': True,
        'aggregationType': 'total',
        'crossSeriesGroupByColumns': time_series_partition_spec.cross_series_group_by_columns
    }

    project_without_target.set_target('target', partitioning_method=time_series_partition_spec)

    assert responses.calls[7].request.method == 'PATCH'
    actual_request = request_body_to_json(responses.calls[7].request)
    assert set(actual_request.keys()) == set(expected_payload.keys())
    assert actual_request == expected_payload


@responses.activate
def test_use_cross_series_group_by_with_invalid_columns(
        time_series_partition_spec, project_with_target_json,
        project_without_target, unittest_endpoint, project_id):

    time_series_partition_spec.multiseries_id_columns = ['series_id']
    time_series_partition_spec.use_cross_series_features = True
    time_series_partition_spec.aggregation_type = 'total'
    time_series_partition_spec.cross_series_group_by_columns = ['location']

    project_url = '{}/projects/{}/'.format(unittest_endpoint, project_id)
    aim_url = project_url + 'aim/'
    status_url = '{}/status/some-status/'.format(unittest_endpoint)
    multi_prop_url = '{}features/{}/multiseriesProperties/'.format(
        project_url, time_series_partition_spec.datetime_partition_column)
    multi_prop_json = {
        'datetimePartitionColumn': time_series_partition_spec.datetime_partition_column,
        'detectedMultiseriesIdColumns': [
            {'multiseriesIdColumns': ['series_id'], 'timeStep': 1, 'timeUnit': 'DAY'}
        ]
    }
    responses.add(responses.GET, multi_prop_url,
                  json=multi_prop_json)

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.datetime_partition_column),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.datetime_partition_column,
                      'id': 2,
                      'timeSeriesEligible': False,
                      'featureType': 'date',
                      'timeSeriesEligibilityReason': 'notUnique',
                      'lowInformation': False, 'uniqueCount': 500
                  })

    responses.add(responses.PATCH,
                  aim_url,
                  body='',
                  status=202,
                  adding_headers={'Location': status_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  status_url,
                  body='',
                  status=303,
                  adding_headers={'Location': project_url},
                  content_type='application/json')

    responses.add(responses.GET,
                  project_url,
                  body=project_with_target_json,
                  status=200,
                  content_type='application/json')

    responses.add(responses.GET,
                  '{}features/{}/'.format(project_url,
                                          time_series_partition_spec.multiseries_id_columns[0]),
                  json={
                      'projectId': project_id,
                      'name': time_series_partition_spec.multiseries_id_columns[0],
                      'id': 3,
                      'timeSeriesEligible': True,
                      'featureType': 'category',
                      'timeSeriesEligibilityReason': 'suitable',
                      'lowInformation': False,
                      'uniqueCount': 100
                  })

    responses.add(responses.GET, ('{}multiseriesIds/{}/crossSeriesProperties/?'
                                  'crossSeriesGroupByColumns={}').format(
        project_url, time_series_partition_spec.multiseries_id_columns[0],
        time_series_partition_spec.cross_series_group_by_columns[0]),
                  json={
                      'multiseriesId': time_series_partition_spec.multiseries_id_columns[0],
                      'crossSeriesGroupByColumns': [
                          {
                              'name': time_series_partition_spec.cross_series_group_by_columns[0],
                              'eligibility': 'splitSeries',
                              'isEligible': False
                          }
                      ]
                  })

    try:
        project_without_target.set_target('target', partitioning_method=time_series_partition_spec)
        assert False, 'Should fail with exception as column \'{}\' is not eligible.'.format(
            time_series_partition_spec.cross_series_group_by_columns[0]
        )
    except InvalidUsageError:
        assert True
